"""This module contains sketching algorithms for reducing dimensionality.

A "sketch" refers to a technique for compressing large matrices into
much smaller representations while approximately preserving their
essential properties.
"""

from abc import ABC, abstractmethod

import numpy as np
from numpy import linalg as la
from numpy.typing import NDArray
from sklearn.utils.extmath import randomized_range_finder


class Sketcher(ABC):  # noqa: PLW1641
    """Base class for sketchers."""

    @abstractmethod
    def left_project(self, M: NDArray[np.float64]) -> NDArray[np.float64]:
        """Define left projector."""

    @abstractmethod
    def right_project(self, M: NDArray[np.float64]) -> NDArray[np.float64]:
        """Define right projector."""

    @abstractmethod
    def set_up_column_sampler(self, A: NDArray[np.float64]) -> None:
        """Set up column sampler for left sketch."""

    @abstractmethod
    def set_up_row_sampler(self, A: NDArray[np.float64]) -> None:
        """Set up row sampler for right sketch."""

    @abstractmethod
    def sample_column_idx(self, rng: np.random.RandomState) -> int:
        """Sample column index of left sketch."""

    @abstractmethod
    def sample_row_idx(self, rng: np.random.RandomState) -> int:
        """Sample row index of right sketch."""

    def __eq__(self, other: object) -> bool:
        """Check for equality for serialization purposes."""
        if not isinstance(other, Sketcher):
            return NotImplemented

        # Get the dictionaries of attributes
        self_attrs = self.__dict__
        other_attrs = other.__dict__

        # Check the keys
        if self_attrs.keys() != other_attrs.keys():
            return False

        # Iterate and compare values
        for key, self_value in self_attrs.items():
            other_value = other_attrs[key]

            # Check if they are numpy arrays and compare accordingly
            is_self_np = isinstance(self_value, np.ndarray)
            is_other_np = isinstance(other_value, np.ndarray)

            if is_self_np and is_other_np:
                if not np.array_equal(self_value, other_value):
                    return False
            elif is_self_np != is_other_np or self_value != other_value:
                return False

        return True


class FKV(Sketcher):
    """FKV (Frieze-Kannan-Vempala) sketching."""

    def __init__(  # noqa: PLR0913
        self,
        A: NDArray[np.float64],
        r: int,
        c: int,
        ls_prob_rows: NDArray[np.float64],
        ls_prob_columns_2d: NDArray[np.float64],
        frobenius: float,
        rng: np.random.RandomState,
    ) -> None:
        """Init :py:class:`FKV`.

        Note: LS stands for length-square.

        Args:
            A: coefficient matrix `A`.
            r: number of rows for left projection matrix.
            c: number of columns for right projection matrix.
            ls_prob_rows: row LS probability distribution of `A`.
            ls_prob_columns_2d: column LS probability distribution of `A` (2D).
            frobenius: Frobenius norm of `A`.
            rng: random state.
        """
        m_rows, n_cols = A.shape

        # Sample row indices
        sampled_rows_idx = rng.choice(m_rows, r, replace=True, p=ls_prob_rows).astype(
            np.uint32
        )

        # Sample column indices
        sampled_columns_idx = np.zeros(c, dtype=np.uint32)
        for j in range(c):
            # Sample row index uniformly at random
            i = rng.choice(sampled_rows_idx, replace=True)

            # Sample column from LS distribution of row `i`
            sampled_columns_idx[j] = rng.choice(n_cols, 1, p=ls_prob_columns_2d[i, :])[
                0
            ]

        # Compute norms
        sampled_row_norms = la.norm(A[sampled_rows_idx, :], axis=1)
        R = (
            A[sampled_rows_idx, :]
            * frobenius
            / (np.sqrt(r) * sampled_row_norms[:, None])
        )
        R_sampled_column_norms = la.norm(R[:, sampled_columns_idx], axis=0)

        # Set sketching parameters
        self._r = r
        self._c = c
        self._sampled_rows_idx = sampled_rows_idx
        self._sampled_columns_idx = sampled_columns_idx
        self._frobenius = frobenius
        self._sampled_row_norms = sampled_row_norms
        self._R_sampled_column_norms = R_sampled_column_norms

    def left_project(self, M: NDArray[np.float64]) -> NDArray[np.float64]:
        """Define left projector."""
        return np.asarray(
            M[self._sampled_rows_idx, :]
            * self._frobenius
            / (np.sqrt(self._r) * self._sampled_row_norms[:, None])
        )

    def right_project(self, M: NDArray[np.float64]) -> NDArray[np.float64]:
        """Define right projector."""
        return np.asarray(
            M[:, self._sampled_columns_idx]
            * self._frobenius
            / (np.sqrt(self._c) * self._R_sampled_column_norms[None, :])
        )

    def set_up_column_sampler(self, A: NDArray[np.float64]) -> None:
        """Build LS distribution to sample columns from matrix `R`."""
        R = self.left_project(A)
        self._R_ls_prob_columns = R**2 / la.norm(R, axis=1)[:, None] ** 2
        self._n_cols = A.shape[1]

    def set_up_row_sampler(self, A: NDArray[np.float64]) -> None:
        """Build LS distribution to sample rows from matrix `C`."""
        C = self.right_project(A)
        self._C_ls_prob_rows = C**2 / la.norm(C, axis=0)[None, :] ** 2
        self._n_rows = A.shape[0]

    def sample_column_idx(self, rng: np.random.RandomState) -> int:
        """Sample a column index from `R`."""
        # Sample row index uniformly at random
        sample_i = rng.choice(self._r)

        # Sample column index from LS distribution of corresponding row
        sample_j = rng.choice(self._n_cols, 1, p=self._R_ls_prob_columns[sample_i, :])[
            0
        ]

        return int(sample_j)

    def sample_row_idx(self, rng: np.random.RandomState) -> int:
        """Sample a row index from `C`."""
        # Sample col index uniformly at random
        sample_j = rng.choice(self._c)

        # Sample row index from LS distribution of corresponding column
        sample_i = rng.choice(self._n_rows, 1, p=self._C_ls_prob_rows[:, sample_j])[0]

        return int(sample_i)


class Halko(Sketcher):
    """Halko sketching."""

    def __init__(
        self,
        A: NDArray[np.float64],
        r: int,
        c: int,
        ls_prob_columns: NDArray[np.float64],
        rng: np.random.RandomState,
    ) -> None:
        """Init :py:class:`Halko`.

        Note: LS stands for length-square.

        Args:
            A: coefficient matrix `A`.
            r: number of rows for left projection matrix.
            c: number of columns for right projection matrix.
            ls_prob_columns: column LS probability distribution of `A`.
            rng: random state.
        """
        self._Q_left = Halko._get_low_dimensional_projector(
            A, axis=0, n_components=r, random_state=rng
        )
        self._Q_right = Halko._get_low_dimensional_projector(
            self._Q_left @ A, axis=1, n_components=c, random_state=rng
        )
        self._ls_prob_columns = ls_prob_columns

    @classmethod
    def _get_low_dimensional_projector(
        cls,
        M: NDArray[np.float64],
        axis: int,
        n_components: int,
        random_state: np.random.RandomState,
    ) -> NDArray[np.float64]:
        """Find random matrix to reduce dimensionality of axis of `M`."""
        n_oversamples = 10
        n_random = n_components + n_oversamples
        n_iter = 7 if n_components < 0.1 * min(M.shape) else 4
        if axis == 1:
            M = M.T
        Q = np.asarray(
            randomized_range_finder(
                M,
                size=n_random,
                n_iter=n_iter,
                power_iteration_normalizer="auto",
                random_state=random_state,
            )
        )
        if axis == 0:
            Q = Q.T

        return np.asarray(Q)

    def left_project(self, M: NDArray[np.float64]) -> NDArray[np.float64]:
        """Define left projector."""
        return self._Q_left @ M

    def right_project(self, M: NDArray[np.float64]) -> NDArray[np.float64]:
        """Define right projector."""
        return M @ self._Q_right

    def set_up_column_sampler(self, A: NDArray[np.float64]) -> None:
        """No setup required."""

    def set_up_row_sampler(self, A: NDArray[np.float64]) -> None:
        """Build LS distribution to sample rows from matrix `C`."""
        A_row_norms = la.norm(A, axis=1)
        A_row_norms_squared = A_row_norms**2
        A_frobenius = np.sqrt(np.sum(A_row_norms_squared))
        self._ls_prob_rows = A_row_norms_squared / A_frobenius**2
        self._n_rows = A.shape[0]

    def sample_column_idx(self, rng: np.random.RandomState) -> int:
        """Sample a column index."""
        n_cols = self._ls_prob_columns.size
        sample_j = rng.choice(n_cols, 1, p=self._ls_prob_columns)[0]

        return int(sample_j)

    def sample_row_idx(self, rng: np.random.RandomState) -> int:
        """Sample a row index."""
        sample_i = rng.choice(self._n_rows, 1, p=self._ls_prob_rows)[0]

        return int(sample_i)



import logging
from typing import SupportsFloat

import numpy as np
import pandas as pd
import pytest
from numpy.linalg import pinv
from numpy.typing import ArrayLike, NDArray

from tno.quantum.ml.components import check_estimator_serializable
from tno.quantum.ml.regression.linear_regression import QILinearEstimator

logging.basicConfig(
    format="%(levelname)s:%(message)s", level=logging.INFO, datefmt="%Y-%m-%d %H:%M:%S"
)


def _compute_n_matches(
    x_reference: ArrayLike,
    expected_top_idx: ArrayLike,
) -> int:
    """Computes the number of matches between `x_reference` and `expected_top_idx`.

    This function determines how many of the indices provided in `expected_top_idx` are
    also present among the indices corresponding to the `k` largest values in
    `x_reference`, where `k` is the number of indices in `expected_top_idx`.

    Args:
        x_reference: a 1-D array-like object of numerical values. The function will
                     consider the absolute values of this array.
        expected_top_idx: a 1-D array-like object of expected top indices.

    Returns:
        The number of indices that are present in both `expected_top_idx` and the top
        `k` indices of `x_reference`.
    """
    # Ensure inputs are numpy arrays for consistency
    x_reference = np.asarray(x_reference)
    expected_top_idx = np.asarray(expected_top_idx)

    # Determine k, the number of top elements to consider
    k = expected_top_idx.size
    if k == 0:
        return 0

    # Get the indices that would sort the absolute values of x_reference
    # in descending order and take the top k.
    top_k_indices = np.flip(np.argsort(np.abs(x_reference)))[:k]

    # Use set intersection to find the number of common indices
    return len(set(top_k_indices) & set(expected_top_idx))


def test_compute_n_matches() -> None:
    """Test computation of the number of matches."""
    x_reference = [0.1, 0.5, 0.2, 0.8]
    expected_top_idx = [1, 3]
    assert _compute_n_matches(x_reference, expected_top_idx) == 2

    x_reference = [10, 5, 2, 8]
    expected_top_idx = [0, 2]  # Expecting indices 0 and 2 to be the top.
    # The actual top 2 indices are 0 and 3 (for values 10 and 8).
    # Only index 0 is a match.
    assert _compute_n_matches(x_reference, expected_top_idx) == 1


def _load_data(
    underdetermined: bool = False,
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Load sample dataset."""
    rng = np.random.RandomState(9)
    rank = 3
    m = 500
    n = 250
    A = rng.normal(0, 1, (m, n))
    U, S, V = np.linalg.svd(A, full_matrices=False)
    S[rank:] = 0
    A = U @ np.diag(S) @ V
    if underdetermined:
        A = A.T
    x = np.asarray(rng.normal(0, 1, A.shape[1]), dtype=np.float64)
    b = A @ x

    return A, b, x


def test_solve_qi() -> None:
    """Test quantum-inspired linear solver."""
    # Load data
    A, b, _ = _load_data()

    # Solve using quantum-inspired algorithm
    rank = 3
    r = 70
    c = 70
    n_samples = 100
    n_entries_x = 10
    rng = 7
    qi = QILinearEstimator(r, c, rank, n_samples, rng)
    qi = qi.fit(A, b)
    sampled_indices, sampled_x = qi.sample_prediction_x(A, n_entries_x)

    print(sampled_indices)  # noqa: T201
    print(sampled_x)  # noqa: T201
    assert np.all(
        sampled_indices == np.asarray([44, 127, 239, 228, 81, 19, 249, 3, 208, 72])
    )
    assert np.allclose(
        sampled_x,
        [
            -0.12064711,
            -0.07698437,
            -0.26102075,
            -0.12162569,
            0.08193343,
            0.14820182,
            -0.16637135,
            0.16646659,
            -0.21766188,
            -0.09149265,
        ],
    )


def test_solve_qi_b() -> None:
    """Test quantum-inspired linear solver to predict `b`."""
    # Load data
    A, b, _ = _load_data()

    # Solve using quantum-inspired algorithm
    rank = 3
    r = 70
    c = 70
    n_samples = 100
    n_entries_b = 10
    rng = 7
    qi = QILinearEstimator(r, c, rank, n_samples, rng)
    qi = qi.fit(A, b)
    sampled_indices, sampled_b = qi.sample_prediction_b(A, n_entries_b)

    print(sampled_indices)  # noqa: T201
    print(sampled_b)  # noqa: T201
    assert np.all(
        sampled_indices == np.asarray([231, 461, 461, 381, 50, 334, 323, 381, 59, 174])
    )
    assert np.allclose(
        sampled_b,
        [
            5.44005314,
            4.49647611,
            4.49647611,
            -4.93110314,
            3.88793886,
            -5.33528434,
            2.32686741,
            -4.93110314,
            3.58986512,
            3.30552512,
        ],
    )


def test_solve_qi_ridge() -> None:
    """Test quantum-inspired ridge regression."""
    # Load data
    A, b, _ = _load_data()

    # Solve using quantum-inspired algorithm
    rank = 3
    r = 70
    c = 70
    n_samples = 100
    n_entries_x = 10
    rng = 7

    def func(arg: SupportsFloat) -> SupportsFloat:
        arg = float(arg)
        return (arg**2 + 0.3) / arg

    qi = QILinearEstimator(r, c, rank, n_samples, rng, func=func)
    qi = qi.fit(A, b)
    sampled_indices, sampled_x = qi.sample_prediction_x(A, n_entries_x)

    print(sampled_indices)  # noqa: T201
    print(sampled_x)  # noqa: T201
    assert np.all(
        sampled_indices == np.asarray([44, 127, 239, 228, 81, 19, 249, 3, 208, 72])
    )
    assert np.allclose(
        sampled_x,
        [
            -0.12061417,
            -0.07696354,
            -0.26095131,
            -0.12159334,
            0.08191188,
            0.14816215,
            -0.16632772,
            0.1664221,
            -0.2176028,
            -0.09146953,
        ],
    )


@pytest.mark.parametrize(
    ("sketcher_name", "random_state", "n_matches_expected"),
    [("fkv", 111, 48), ("halko", 11, 49)],
)
def test_finding_largest_entries_b_underdetermined(
    sketcher_name: str, random_state: int, n_matches_expected: int
) -> None:
    """Test quantum-inspired least squares."""
    # Load data
    A, b, _ = _load_data(underdetermined=True)
    top_size = 50

    # Solve using quantum-inspired algorithm
    rank = 3
    r = 70
    c = 80
    n_samples = 100
    n_entries_b = 1000
    rng = random_state
    qi = QILinearEstimator(r, c, rank, n_samples, rng, sketcher_name=sketcher_name)
    qi = qi.fit(A, b)
    sampled_indices, sampled_b = qi.sample_prediction_b(A, n_entries_b)

    # Compare results
    df = pd.DataFrame({"b_idx_samples": sampled_indices, "b_samples": sampled_b})  # noqa: PD901
    df_mean = df.groupby("b_idx_samples")["b_samples"].mean()
    unique_sampled_indices = np.asarray(df_mean.keys())
    unique_sampled_b = np.asarray(df_mean.values)
    sort_idx = np.flip(np.argsort(np.abs(unique_sampled_b)))
    b_idx = unique_sampled_indices[sort_idx][:top_size]
    n_matches = _compute_n_matches(b, b_idx)

    assert n_matches == n_matches_expected


def test_finding_largest_entries_b() -> None:
    """Test quantum-inspired least squares."""
    # Load data
    A, b, _ = _load_data()
    top_size = 25

    # Solve using quantum-inspired algorithm
    rank = 3
    r = 70
    c = 80
    n_samples = 100
    n_entries_b = 1000
    rng = 111
    qi = QILinearEstimator(r, c, rank, n_samples, rng)
    qi = qi.fit(A, b)
    sampled_indices, sampled_b = qi.sample_prediction_b(A, n_entries_b)

    # Compare results
    df = pd.DataFrame({"b_idx_samples": sampled_indices, "b_samples": sampled_b})  # noqa: PD901
    df_mean = df.groupby("b_idx_samples")["b_samples"].mean()
    unique_sampled_indices = np.asarray(df_mean.keys())
    unique_sampled_b = np.asarray(df_mean.values)
    sort_idx = np.flip(np.argsort(np.abs(unique_sampled_b)))
    b_idx = unique_sampled_indices[sort_idx][:top_size]
    n_matches = _compute_n_matches(b, b_idx)

    assert n_matches == 23


def test_pseudoinverse() -> None:
    """Test pseudoinverse."""
    # Load data
    A, b, _ = _load_data()
    rank = 3

    # Solve using pseudoinverse
    x_sol = pinv(A).dot(b)

    # Solve using pseudoinverse II
    U, S, V = np.linalg.svd(A, full_matrices=False)
    A_pinv = V[:rank, :].T @ (np.diag(1 / S[:rank])) @ U[:, :rank].T
    x_sol2 = A_pinv @ b

    # Solve using pseudoinverse III
    sigmas = S[:rank]
    lambdas = []
    for ell in range(rank):
        lambdas.append(1 / (sigmas[ell]) ** 2 * np.sum(A * (np.outer(b, V[ell, :]))))  # noqa: PERF401
    x_sol3 = np.squeeze(
        A.T @ (U[:, :rank] @ (np.asarray(lambdas)[:, None] / sigmas[:, None]))
    )

    assert np.allclose(x_sol, x_sol2)
    assert np.allclose(x_sol, x_sol3)


def test_finding_largest_entries_x() -> None:
    """Test quantum-inspired least squares."""
    # Load data
    A, b, _ = _load_data()
    top_size = 25

    # Solve using quantum-inspired algorithm
    rank = 3
    r = 70
    c = 70
    n_samples = 100
    n_entries_x = 1000
    rng = 7
    qi = QILinearEstimator(r, c, rank, n_samples, rng)
    qi = qi.fit(A, b)
    sampled_indices, sampled_x = qi.sample_prediction_x(A, n_entries_x)

    # Solve using pseudoinverse
    x_sol = pinv(A).dot(b)

    # Compare results
    df = pd.DataFrame({"x_idx_samples": sampled_indices, "x_samples": sampled_x})  # noqa: PD901
    df_mean = df.groupby("x_idx_samples")["x_samples"].mean()
    unique_sampled_indices = np.asarray(df_mean.keys())
    unique_sampled_x = np.asarray(df_mean.values)
    sort_idx = np.flip(np.argsort(np.abs(unique_sampled_x)))
    x_idx = unique_sampled_indices[sort_idx][:top_size]
    n_matches = _compute_n_matches(x_sol, x_idx)

    assert n_matches == 23


@pytest.mark.parametrize(
    "sketcher_name",
    ["fkv", "halko"],
)
def test_serialization(sketcher_name: str) -> None:
    """Test correct serialization of estimator."""
    # Load data
    A, b, _ = _load_data()

    # Define parameters
    rank = 3
    r = 70
    c = 70
    n_samples = 100
    rng = 7

    # Check serialization
    qi1 = QILinearEstimator(r, c, rank, n_samples, rng, sketcher_name=sketcher_name)
    check_estimator_serializable(qi1)

    qi2 = QILinearEstimator(r, c, rank, n_samples, rng, sketcher_name=sketcher_name)
    qi2 = qi2.fit(A, b)
    check_estimator_serializable(qi2)


if __name__ == "__main__":
    test_serialization(sketcher_name="fkv")



import numpy as np
import pytest  # noqa: F401
from numpy.typing import NDArray

from tno.quantum.ml.regression.linear_regression._quantum_inspired import (
    compute_ls_probs,
)
from tno.quantum.ml.regression.linear_regression._sketching import FKV, Halko


def _get_FKV_sketcher(A: NDArray[np.float64], r: int, c: int) -> FKV:  # noqa: N802
    """Load dummy sketcher."""
    A_ls_prob_rows, A_ls_prob_columns_2d, _, _, A_frobenius = compute_ls_probs(A)
    r = 30
    c = 40
    random_state = np.random.RandomState(7)
    return FKV(
        A,
        r,
        c,
        A_ls_prob_rows,
        A_ls_prob_columns_2d,
        A_frobenius,
        random_state,
    )


def _get_Halko_sketcher(A: NDArray[np.float64], r: int, c: int) -> Halko:  # noqa: N802
    """Load dummy sketcher."""
    _, _, A_ls_prob_columns, _, _ = compute_ls_probs(A)
    random_state = np.random.RandomState(7)
    return Halko(
        A,
        r,
        c,
        A_ls_prob_columns,
        random_state,
    )


def test_FKV_dimensions() -> None:  # noqa: N802
    """Test dimensions of sketches."""
    A = np.arange(100 * 100, dtype=np.float64).reshape((100, 100))
    r = 30
    c = 40
    sketcher = _get_FKV_sketcher(A, r, c)

    left_sketch_matrix = sketcher.left_project(A)
    right_sketch_matrix = sketcher.right_project(A)
    left_right_sketch_matrix = sketcher.right_project(sketcher.left_project(A))

    assert left_sketch_matrix.shape == (30, 100)
    assert right_sketch_matrix.shape == (100, 40)
    assert left_right_sketch_matrix.shape == (30, 40)


def test_Halko_dimensions() -> None:  # noqa: N802
    """Test dimensions of sketches."""
    A = np.arange(100 * 100, dtype=np.float64).reshape((100, 100))
    r = 30
    c = 40
    sketcher = _get_Halko_sketcher(A, r, c)

    left_sketch_matrix = sketcher.left_project(A)
    right_sketch_matrix = sketcher.right_project(A)
    left_right_sketch_matrix = sketcher.right_project(sketcher.left_project(A))

    assert left_sketch_matrix.shape == (r + 10, 100)
    assert right_sketch_matrix.shape == (100, r + 10)
    assert left_right_sketch_matrix.shape == (r + 10, r + 10)


def test_FKV_samplers() -> None:  # noqa: N802
    """Test FKV-based samplers."""
    A = np.arange(100 * 100, dtype=np.float64).reshape((100, 100))
    r = 30
    c = 40
    A_predict = np.arange(150 * 100, dtype=np.float64).reshape((150, 100))
    sketcher = _get_FKV_sketcher(A, r, c)

    sketcher.set_up_column_sampler(A)
    assert np.all(np.allclose(np.sum(sketcher._R_ls_prob_columns, axis=1), 1))
    assert sketcher._R_ls_prob_columns.shape == (r, A.shape[1])
    assert sketcher._n_cols == A.shape[1]

    sketcher.set_up_row_sampler(A_predict)
    assert np.all(np.allclose(np.sum(sketcher._C_ls_prob_rows, axis=0), 1))
    assert sketcher._C_ls_prob_rows.shape == (A_predict.shape[0], c)
    assert sketcher._n_rows == A_predict.shape[0]



"""This module implements a quantum-inspired algorithm for linear regression.

Assume a linear system of the form `$Ax=b$` where:

- `$A$` is the training data.
- `$x$` is a vector of unknown coefficients.
- `$b$` is a vector of target values.

The class :py:class:`QILinearEstimator` provides three methods:

- :py:meth:`~QILinearEstimator.fit`: for model fitting using `$A$` and `$b$`.
- :py:meth:`~QILinearEstimator.sample_prediction_x`: for sampling entries of
  the estimated coefficient vector.
- :py:meth:`~QILinearEstimator.sample_prediction_b`: for sampling entries of predictions
  corresponding to (un)observed target values.


Basic usage
-----------

>>> import numpy as np
>>> from sklearn.datasets import make_low_rank_matrix
>>> from sklearn.model_selection import train_test_split
>>> from tno.quantum.ml.regression.linear_regression import QILinearEstimator
>>>
>>> rng = np.random.RandomState(7)
>>>
>>> # Generate example data
>>> m = 700
>>> n = 100
>>> A = make_low_rank_matrix(n_samples=m, n_features=n, effective_rank=3,
...     random_state=rng, tail_strength=0.1)
>>> x = rng.normal(0, 1, A.shape[1])
>>> b = A @ x
>>>
>>> # Create training and test datasets
>>> A_train, A_test, b_train, b_test = train_test_split(A, b, test_size=0.3,
...     random_state=rng)
>>>
>>> # Fit quantum-inspired model
>>> rank = 3
>>> r = 100
>>> c = 30
>>> n_samples = 100  # for Monte Carlo methods
>>> qi = QILinearEstimator(r, c, rank, n_samples, rng, sketcher_name="fkv")
>>> qi = qi.fit(A_train, b_train)
>>>
>>> # Sample from b (vector of predictions)
>>> n_entries_b = 1000
>>> sampled_indices_b, sampled_b = qi.sample_prediction_b(A_test, n_entries_b)
"""

from tno.quantum.ml.regression.linear_regression._estimator import QILinearEstimator

__all__: list[str] = ["QILinearEstimator"]

__version__ = "1.0.0"



import numpy as np
import pytest  # noqa: F401

from tno.quantum.ml.regression.linear_regression._quantum_inspired import (
    compute_ls_probs,
)


def test_compute_ls_probs() -> None:
    """Test LS probabilities."""
    A = np.arange(100, dtype=np.float64).reshape((10, 10))

    A_ls_prob_rows, A_ls_prob_columns_2d, A_ls_prob_columns, A_ls_prob_rows_2d, _ = (
        compute_ls_probs(A)
    )

    tolerance = -1e-8

    assert np.allclose(np.sum(A_ls_prob_rows), 1)
    assert np.all(np.allclose(np.sum(A_ls_prob_columns_2d, axis=1), 1))
    assert np.all(A_ls_prob_rows > tolerance)
    assert np.all(A_ls_prob_columns_2d > tolerance)

    assert np.allclose(np.sum(A_ls_prob_columns), 1)
    assert np.all(np.allclose(np.sum(A_ls_prob_rows_2d, axis=0), 1))
    assert np.all(A_ls_prob_columns > tolerance)
    assert np.all(A_ls_prob_rows_2d > tolerance)


if __name__ == "__main__":
    test_compute_ls_probs()



"""Module with definitions of quantum-inspired estimators."""

from __future__ import annotations

import logging
import warnings
from typing import Any, Callable, SupportsFloat, SupportsInt

import numpy as np
from numpy import linalg as la
from numpy.typing import ArrayLike, NDArray

from tno.quantum.ml.components import SerializableEstimator
from tno.quantum.ml.regression.linear_regression._quantum_inspired import (
    compute_ls_probs,
    estimate_lambdas,
    sample_from_b,
    sample_from_x,
)
from tno.quantum.ml.regression.linear_regression._sketching import FKV, Halko, Sketcher
from tno.quantum.utils.serialization import Serializable
from tno.quantum.utils.validation import (
    check_arraylike,
    check_int,
    check_random_state,
    check_real,
)

logger = logging.getLogger(__name__)


def _serialize_sketcher(sketcher: Sketcher) -> dict[str, Any]:
    dict_ = {}
    for key, value in sketcher.__dict__.items():
        dict_[key] = Serializable.serialize(value)
    return dict_


def _deserialize_sketcher(
    data: dict[str, Any], sketcher_name: str = "fkv"
) -> FKV | Halko:
    data = {key: Serializable.deserialize(value) for key, value in data.items()}
    sketcher_class: type[FKV | Halko]
    if sketcher_name == "fkv":
        sketcher_class = FKV
    elif sketcher_name == "halko":
        sketcher_class = Halko
    else:
        message = '`sketcher_name` should be either "fkv" or "halko"'
        raise ValueError(message)
    sketcher = sketcher_class.__new__(sketcher_class)
    sketcher.__dict__.update(data)
    return sketcher


Serializable.register(
    FKV, _serialize_sketcher, lambda x: _deserialize_sketcher(x, sketcher_name="fkv")
)
Serializable.register(
    Halko,
    _serialize_sketcher,
    lambda x: _deserialize_sketcher(x, sketcher_name="halko"),
)


class EstimatorError(Exception):
    """Module exception."""

    def __init__(self, message: str) -> None:
        """Init :py:class:`EstimatorError`."""
        super().__init__(message)


class QILinearEstimator(SerializableEstimator):  # noqa: PLW1641
    """Quantum-inspired linear estimator."""

    def __init__(  # noqa: PLR0913
        self,
        r: SupportsInt,
        c: SupportsInt,
        rank: SupportsInt,
        n_samples: SupportsInt,
        random_state: SupportsInt | None = None,
        sigma_threshold: SupportsFloat = 1e-15,
        sketcher_name: str = "fkv",
        func: Callable[[SupportsFloat], SupportsFloat] | None = None,
    ) -> None:
        """Init :py:class:`QILinearEstimator`.

        Args:
            r: number of rows to sample from `A`.
            c: number of columns to sample from `A`.
            rank: rank used to approximate matrix `A`.
            n_samples: number of samples to estimate inner products.
                       Note: the sampling is  performed from entries of `A`,
                       so there are ``A.shape[0] * A.shape[1]`` possible entries.
            random_state: random seed.
            sigma_threshold: the argument `rank` is recomputed in case it is higher
                             the number of singular values below this threhold.
            sketcher_name: name of sketching method: ``"fkv"`` or ``"halko"``.
            func: function to transform singular values when estimating lambda
                  coefficients. This can be used for Tikhonov regularization purposes.
        """
        self.r = r
        self.c = c
        self.rank = rank
        self.n_samples = n_samples
        self.random_state = random_state
        self.sigma_threshold = sigma_threshold
        self.sketcher_name = sketcher_name
        self.func = func

    def fit(
        self,
        A: ArrayLike,
        b: ArrayLike,
    ) -> QILinearEstimator:
        """Fit data using quantum-inspired algorithm.

        Args:
            A: coefficient matrix `A`.
            b: vector `b`.
        """
        # Validate input
        A = check_arraylike(A, "A", ndim=2)
        A = np.asarray(A, dtype=np.float64)
        b = check_arraylike(b, "b", ndim=1)
        b = np.asarray(b, dtype=np.float64)
        self.rank = check_int(self.rank, "rank", l_bound=1, l_inclusive=True)
        self.r = check_int(self.r, "r", l_bound=self.rank, l_inclusive=False)
        self.c = check_int(self.c, "c", l_bound=self.rank, l_inclusive=False)
        self.n_samples = check_int(
            self.n_samples, "n_samples ", l_bound=2, l_inclusive=True
        )
        self.sigma_threshold = check_real(
            self.sigma_threshold, "sigma_threshold", l_bound=0, l_inclusive=False
        )

        # Get random state
        rng = check_random_state(self.random_state, "random_state")

        # 1. Generate length-square probability distributions to sample from matrix `A`
        logger.info(
            "1. Generate length-square probability distributions to sample "
            "from matrix `A`"
        )
        (
            A_ls_prob_rows,
            A_ls_prob_columns_2d,
            A_ls_prob_columns,
            _,
            A_frobenius,
        ) = compute_ls_probs(A)

        # 2. Build matrix `C`
        logger.info("2. Build matrix `C`")
        self.sketcher_: Sketcher
        if self.sketcher_name == "fkv":
            self.sketcher_ = FKV(
                A,
                self.r,
                self.c,
                A_ls_prob_rows,
                A_ls_prob_columns_2d,
                A_frobenius,
                rng,
            )
        elif self.sketcher_name == "halko":
            self.sketcher_ = Halko(
                A,
                self.r,
                self.c,
                A_ls_prob_columns,
                rng,
            )
        else:
            message = '`sketcher_name` should be either "fkv" or "halko"'
            raise ValueError(message)

        C = self.sketcher_.right_project(self.sketcher_.left_project(A))

        # 3. Compute the SVD of `C`
        logger.info("3. Compute the SVD of `C`")
        self.w_left_, self.sigma_, self.w_right_T_ = la.svd(C, full_matrices=False)

        # Recompute rank
        self.rank_ = self.rank
        rank_recomputed = int(np.count_nonzero(self.sigma_ > self.sigma_threshold))
        if rank_recomputed < self.rank:
            message = f"Desired rank: {self.rank}; recomputed: {rank_recomputed}"
            warnings.warn(message, RuntimeWarning, stacklevel=2)
            logger.warning(message)
            self.rank_ = rank_recomputed

        # 4. Estimate lambda coefficients
        logger.info("4. Estimate lambda coefficients")
        func: Callable[[SupportsFloat], SupportsFloat]
        if self.func is None:

            def func_(arg: SupportsFloat) -> SupportsFloat:
                return arg

            func = func_
        else:
            func = self.func
        self.lambdas_ = estimate_lambdas(
            A,
            b,
            self.n_samples,
            self.rank_,
            self.w_left_,
            self.sigma_,
            self.sketcher_,
            A_ls_prob_rows,
            A_ls_prob_columns_2d,
            A_frobenius,
            rng,
            func,
        )

        return self

    def _check_is_fitted(self) -> None:
        """Check if the `fit` method has been called."""
        for attribute_name in [
            "sketcher_",
            "w_left_",
            "sigma_",
            "w_right_T_",
            "rank_",
            "lambdas_",
        ]:
            if not hasattr(self, attribute_name):
                message = "Please call `fit` before making predictions"
                raise EstimatorError(message)

    def sample_prediction_x(
        self,
        A: ArrayLike,
        n_entries_x: SupportsInt,
    ) -> tuple[NDArray[np.uint32], NDArray[np.float64]]:
        """Samples predictions of `x` using quantum-inspired model.

        Args:
            A: coefficient matrix `A`.
            n_entries_x: number of entries to be sampled from the solution vector `x`.
                         Set this to 0 to skip this sampling step.

        Returns:
            Samples of predicted values and corresponding indices.
        """
        self._check_is_fitted()
        rng = check_random_state(self.random_state, "random_state")

        A = check_arraylike(A, "A", ndim=2)
        A = np.asarray(A, dtype=np.float64)
        n_entries_x = check_int(n_entries_x, "n_entries_x", l_bound=1, l_inclusive=True)

        logger.info("Sample predicted `x`")

        # Compute `omega`
        omega = self.w_left_[:, : self.rank_] @ (
            self.lambdas_ / self.sigma_[: self.rank_]
        )
        omega_norm = float(la.norm(omega))

        # Sample entries of solution vector `x`
        sampled_indices_x = np.zeros(n_entries_x, dtype=np.uint32)
        sampled_x = np.zeros(n_entries_x)
        for t in range(n_entries_x):
            sampled_indices_x[t], sampled_x[t] = sample_from_x(
                A,
                self.sketcher_,
                omega,
                omega_norm,
                rng,
            )
            if (t + 1) % 100 == 0:
                logger.info("---%s entries sampled out of %s", t + 1, n_entries_x)

        return sampled_indices_x, sampled_x

    def sample_prediction_b(
        self,
        A: ArrayLike,
        n_entries_b: SupportsInt,
    ) -> tuple[NDArray[np.uint32], NDArray[np.float64]]:
        """Sample predictions of `b` using quantum-inspired model.

        Args:
            A: coefficient matrix `A`.
            n_entries_b: number of entries to be sampled from the predicted `b`.

        Returns:
            Samples of predicted values and corresponding indices.
        """
        self._check_is_fitted()
        rng = check_random_state(self.random_state, "random_state")

        A = check_arraylike(A, "A", ndim=2)
        A = np.asarray(A, dtype=np.float64)
        n_entries_b = check_int(n_entries_b, "n_entries_b", l_bound=1, l_inclusive=True)

        logger.info("Sample predicted `b`")

        # Compute `phi`
        phi = self.w_right_T_.T[:, : self.rank_] @ self.lambdas_
        phi_norm = float(la.norm(phi))

        # Sample entries of `b`
        sampled_indices_b = np.zeros(n_entries_b, dtype=np.uint32)
        sampled_b = np.zeros(n_entries_b)
        for t in range(n_entries_b):
            sampled_indices_b[t], sampled_b[t] = sample_from_b(
                A,
                self.sketcher_,
                phi,
                phi_norm,
                rng,
            )
            if (t + 1) % 100 == 0:
                logger.info("---%s entries sampled out of %s", t + 1, n_entries_b)

        return sampled_indices_b, sampled_b

    def __eq__(self, other: object) -> bool:
        """Check for equality for serialization purposes."""
        if not isinstance(other, QILinearEstimator):
            return NotImplemented

        # Get the dictionaries of attributes
        self_attrs = self.__dict__
        other_attrs = other.__dict__

        # Check the keys
        if set(self_attrs.keys()) != set(other_attrs.keys()):
            return False

        # Iterate and compare values
        for key, self_value in self_attrs.items():
            other_value = other_attrs[key]

            # Check if they are numpy arrays and compare accordingly
            is_self_np = isinstance(self_value, np.ndarray)
            is_other_np = isinstance(other_value, np.ndarray)

            if is_self_np and is_other_np:
                if not np.array_equal(self_value, other_value):
                    return False
            elif is_self_np != is_other_np or self_value != other_value:
                return False

        return True



"""Quantum-inspired sampling utilities for linear regression.

This module provides functions for quantum-inspired sampling techniques used in
in the context of randomized numerical linear algebra and sketching methods.
It includes utilities for computing length-square (LS) probability distributions,
estimating lambda coefficients via Monte Carlo methods,
and performing LS sampling from both coefficient vectors and predictions.
"""

import logging
import warnings
from typing import Callable, SupportsFloat

import numpy as np
from numpy import linalg as la
from numpy.typing import NDArray

from tno.quantum.ml.regression.linear_regression._sketching import Sketcher

logger = logging.getLogger(__name__)


def compute_ls_probs(
    A: NDArray[np.float64],
) -> tuple[
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    NDArray[np.float64],
    float,
]:
    """Compute length-square (LS) probability distributions for sampling `A`.

    Args:
        A: coefficient matrix `A`.

    Returns:
        LS probability distribution for rows,
        LS probability distribution for columns (2D),
        LS probability distribution for columns,
        LS probability distribution for rows (2D),
        Frobenius norm
    """
    # Compute row norms
    A_row_norms = la.norm(A, axis=1)
    A_row_norms_squared = A_row_norms**2

    # Compute column norms
    A_column_norms = la.norm(A, axis=0)
    A_column_norms_squared = A_column_norms**2

    # Compute Frobenius norm
    A_frobenius = float(np.sqrt(np.sum(A_row_norms_squared)))

    # Compute LS probabilities for rows
    A_ls_prob_rows = A_row_norms_squared / A_frobenius**2

    # Compute LS probabilities for columns
    A_ls_prob_columns_2d = A**2 / A_row_norms_squared[:, None]

    # Compute LS probabilities for columns
    A_ls_prob_columns = A_column_norms_squared / A_frobenius**2

    # Compute LS probabilities for rows
    A_ls_prob_rows_2d = A**2 / A_column_norms_squared[None, :]

    return (
        A_ls_prob_rows,
        A_ls_prob_columns_2d,
        A_ls_prob_columns,
        A_ls_prob_rows_2d,
        A_frobenius,
    )


def estimate_lambdas(  # noqa: PLR0913
    A: NDArray[np.float64],
    b: NDArray[np.float64],
    n_samples: int,
    rank: int,
    w: NDArray[np.float64],
    sigma: NDArray[np.float64],
    sketcher: Sketcher,
    A_ls_prob_rows: NDArray[np.float64],
    A_ls_prob_columns: NDArray[np.float64],
    A_frobenius: float,
    rng: np.random.RandomState,
    func: Callable[[SupportsFloat], SupportsFloat],
) -> NDArray[np.float64]:
    """Estimate lambda coefficients.

    Args:
        A: coefficient matrix `A`.
        b: vector `b`.
        n_samples: number of samples to estimate inner products.
                   Note: the sampling is  performed from entries of `A`,
                   so there are ``A.shape[0] * A.shape[1]`` possible entries.
        rank: rank used to approximate matrix `A`.
        w: left-singular vector of `C`.
        sigma: singular values of `C`.
        sketcher: sketcher to left project `A`.
        A_ls_prob_rows: row LS probability distribution of `A`.
        A_ls_prob_columns: column LS probability distribution of `A` (2D).
        A_frobenius: Frobenius norm of `A`.
        rng: random state.
        func: function to transform singular values when estimating lambda coefficients.

    Returns:
        lambda coefficients
    """
    m_rows, n_cols = A.shape
    n_realizations = 10
    lambdas_realizations = np.zeros((n_realizations, rank))
    for realization_i in range(n_realizations):
        logger.info("---Realization %s", realization_i)
        for ell in range(rank):
            # 1. Generate sample indices
            samples_i = []
            samples_j = []
            for _ in range(n_samples):
                i = rng.choice(m_rows, 1, replace=True, p=A_ls_prob_rows)[0]
                j = rng.choice(n_cols, 1, p=A_ls_prob_columns[i])[0]
                samples_i.append(i)
                samples_j.append(j)

            # 2. Approximate lambda using Monte Carlo estimation

            # Estimate right-singular vector
            R = sketcher.left_project(A[:, np.asarray(samples_j)])
            v_approx = R.T @ (w[:, ell] / sigma[ell])

            # Compute entries of outer product between `b` and `v_approx`
            outer_prod_b_v = np.squeeze(b[samples_i]) * v_approx

            # Estimate inner product between `A` and `outer_prod_b_v`
            inner_prod = np.mean(
                A_frobenius**2 / A[samples_i, samples_j] * outer_prod_b_v
            )

            # Compute lambda
            lambdas_realizations[realization_i, ell] = (
                inner_prod / sigma[ell] / func(sigma[ell])
            )

    return np.asarray(np.median(lambdas_realizations, axis=0))


def sample_from_b(  # noqa: PLR0913
    A: NDArray[np.float64],
    sketcher: Sketcher,
    phi: NDArray[np.float64],
    phi_norm: float,
    rng: np.random.RandomState,
    max_n_sampling_attempts: int = int(1e5),
) -> tuple[int, float]:
    """Perform length-square (LS) sampling from the predicted `b`.

    Args:
        A: coefficient matrix `A`.
        sketcher: sketcher to left project `A` and sample its rows.
        phi: vector phi.
        phi_norm: norm of `phi`.
        rng: random state.
        max_n_sampling_attempts: maximum number of sampling attempts.

    Returns:
        index of the sampled entry,
        entry value
    """
    C = sketcher.right_project(A)
    sketcher.set_up_row_sampler(A)

    for _ in range(max_n_sampling_attempts):
        # Sample row index
        sample_i = sketcher.sample_row_idx(rng)

        # Sample row of `C`
        C_i = C[sample_i, :]
        C_i_norm = la.norm(C_i)

        # Determine if we output `sample_i`
        dot_prod_C_i_omega = np.dot(C_i, phi)
        prob = (dot_prod_C_i_omega / (phi_norm * C_i_norm)) ** 2
        if rng.binomial(1, prob) == 1:
            return sample_i, dot_prod_C_i_omega

    message = (
        "Maximum number of sampling attempts "
        f"({max_n_sampling_attempts}) exceeded. Returning sample from last attempt."
    )
    warnings.warn(message, RuntimeWarning, stacklevel=2)
    logger.warning(message)

    return sample_i, dot_prod_C_i_omega


def sample_from_x(  # noqa: PLR0913
    A: NDArray[np.float64],
    sketcher: Sketcher,
    omega: NDArray[np.float64],
    omega_norm: float,
    rng: np.random.RandomState,
    max_n_sampling_attempts: int = int(1e5),
) -> tuple[int, float]:
    """Perform length-square (LS) sampling from the solution vector.

    Args:
        A: coefficient matrix `A`.
        sketcher: sketcher to left project `A` and sample its columns.
        omega: vector `omega`.
        omega_norm: norm of `omega`.
        rng: random state.
        max_n_sampling_attempts: maximum number of sampling attempts.

    Returns:
        index of the sampled entry,
        entry value
    """
    R = sketcher.left_project(A)
    sketcher.set_up_column_sampler(A)

    for _ in range(max_n_sampling_attempts):
        # Sample column index
        sample_j = sketcher.sample_column_idx(rng)

        # Sample column of `R`
        R_j = R[:, sample_j]
        R_j_norm = la.norm(R_j)

        # Determine if we output `sample_j`
        dot_prod_R_j_omega = np.dot(R_j, omega)
        prob = (dot_prod_R_j_omega / (omega_norm * R_j_norm)) ** 2
        if rng.binomial(1, prob) == 1:
            return sample_j, dot_prod_R_j_omega

    message = (
        "Maximum number of sampling attempts "
        f"({max_n_sampling_attempts}) exceeded. Returning sample from last attempt."
    )
    warnings.warn(message, RuntimeWarning, stacklevel=2)
    logger.warning(message)

    return sample_j, dot_prod_R_j_omega
